import torch
import torch.nn as nn


class DotPredictor(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, first_vector, second_vector):
        '''
        Args:
            first_vector: (shape) batch_size, *,  dim
            second_vector: (shape) batch_size, *, dim
        Returns:
            (shape) batch_size, *
        '''
        return torch.mul(first_vector, second_vector).sum(dim=-1)


if __name__ == '__main__':
    first_vector = torch.rand(16, 10)
    second_vector = torch.rand(16, 10)
    dot_predictor = DotPredictor()
    print(dot_predictor(first_vector, second_vector))



import torch
import torch.nn as nn


class DNNPredictor(nn.Module):
    def __init__(self, dims):
        super().__init__()
        assert len(dims) >= 2 and dims[-1] == 1

        layers = []
        for i in range(len(dims) - 2):
            layers.append(nn.Linear(dims[i], dims[i + 1]))
            layers.append(nn.ReLU())
        layers.append(nn.Linear(dims[-2], dims[-1]))

        self.dnn = nn.Sequential(*layers)

    def forward(self, first_vector, second_vector):
        '''
        Args:
            first_vector: (shape) batch_size, *,  dim
            second_vector: (shape) batch_size, *, dim
        Returns:
            (shape) batch_size, *
        '''
        return self.dnn(torch.cat((first_vector, second_vector),
                                  dim=-1)).squeeze(dim=-1)


if __name__ == '__main__':
    first_vector = torch.rand(16, 8, 10)
    second_vector = torch.rand(16, 8, 10)
    dnn_predictor = DNNPredictor([20, 4, 1])
    print(dnn_predictor)
    print(dnn_predictor(first_vector, second_vector))



import torch
import torch.nn as nn
import torch.nn.functional as F

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


class AdditiveAttention(nn.Module):
    """
    A general additive attention module.
    """
    def __init__(self, attention_query_vector_dim, candidate_vector_dim):
        super().__init__()
        self.linear = nn.Linear(candidate_vector_dim,
                                attention_query_vector_dim)
        self.attention_query_vector = nn.Parameter(
            torch.empty(attention_query_vector_dim).uniform_(-0.1, 0.1))

    def forward(self, candidate_vector, mask=None):
        """
        Args:
            candidate_vector: batch_size, candidate_size, candidate_vector_dim
            mask: batch_size, candidate_size
        Returns:
            (shape) batch_size, candidate_vector_dim
        """
        # batch_size, candidate_size, attention_query_vector_dim
        temp = torch.tanh(self.linear(candidate_vector))
        # batch_size, candidate_size
        weights = torch.matmul(temp, self.attention_query_vector)
        if mask is not None:
            weights[~mask] = float('-inf')
        weights = F.softmax(weights, dim=1)
        # batch_size, candidate_vector_dim
        target = torch.bmm(weights.unsqueeze(dim=1),
                           candidate_vector).squeeze(dim=1)
        return target
